from __future__ import unicode_literals

import unittest

import bt_manager
import mock
import dbus
import os


class MockDBusInterface:
    """Mock dbus.Interface implementation for purpose of testing"""
    def __init__(self, obj, addr):
        self.addr = addr
        if (self.addr == 'org.bluez.Adapter'):
            self._props = dbus.Dictionary({dbus.String(u'Name'): dbus.String(u'new-name', variant_level=1),  # noqa
                                           dbus.String(u'Powered'): dbus.Boolean(True, variant_level=1),  # noqa
                                           dbus.String(u'Devices'):
                                           dbus.Array([dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')],  # noqa
                                                      signature=dbus.Signature('o'), variant_level=1),  # noqa
                                           dbus.String(u'DiscoverableTimeout'): dbus.UInt32(0L, variant_level=1),  # noqa
                                           dbus.String(u'PairableTimeout'): dbus.UInt32(0L, variant_level=1),  # noqa
                                           dbus.String(u'Discoverable'): dbus.Boolean(True, variant_level=1),  # noqa
                                           dbus.String(u'Address'): dbus.String(u'AC:7B:A1:3C:13:82', variant_level=1),  # noqa
                                           dbus.String(u'Discovering'): dbus.Boolean(False, variant_level=1),  # noqa
                                           dbus.String(u'Pairable'): dbus.Boolean(True, variant_level=1),  # noqa
                                           dbus.String(u'Class'): dbus.UInt32(7209216L, variant_level=1),  # noqa
                                           dbus.String(u'UUIDs'):
                                           dbus.Array([dbus.String(u'00001000-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'00001001-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000112d-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'00001112-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000111f-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000111e-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000110a-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000110b-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000110c-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'0000110e-0000-1000-8000-00805f9b34fb'),  # noqa
                                                       dbus.String(u'00001103-0000-1000-8000-00805f9b34fb')],  # noqa
                                                      signature=dbus.Signature('s'), variant_level=1)},  # noqa
                                          signature=dbus.Signature('sv'))
        elif (self.addr == 'org.bluez.Device'):
            self._props = dbus.Dictionary({dbus.String(u'Product'): dbus.UInt16(5028, variant_level=1),  # noqa
                                           dbus.String(u'Vendor'): dbus.UInt16(57, variant_level=1),  # noqa
                                           dbus.String(u'Name'): dbus.String(u'BTS-06', variant_level=1),  # noqa
                                           dbus.String(u'Paired'): dbus.Boolean(True, variant_level=1),  # noqa
                                           dbus.String(u'Adapter'): dbus.ObjectPath('/org/bluez/985/hci0',  # noqa
                                                                                    variant_level=1),  # noqa
                                           dbus.String(u'Alias'): dbus.String(u'BTS-06', variant_level=1),  # noqa
                                           dbus.String(u'Version'): dbus.UInt16(260, variant_level=1),  # noqa
                                           dbus.String(u'Connected'): dbus.Boolean(False, variant_level=1),  # noqa
                                           dbus.String(u'UUIDs'):
                                                dbus.Array([dbus.String(u'00001108-0000-1000-8000-00805f9b34fb'),  # noqa
                                                            dbus.String(u'0000110b-0000-1000-8000-00805f9b34fb'),  # noqa
                                                            dbus.String(u'0000110c-0000-1000-8000-00805f9b34fb'),  # noqa
                                                            dbus.String(u'0000110e-0000-1000-8000-00805f9b34fb'),  # noqa
                                                            dbus.String(u'0000111e-0000-1000-8000-00805f9b34fb'),  # noqa
                                                            dbus.String(u'00001200-0000-1000-8000-00805f9b34fb')],  # noqa
                                                           signature=dbus.Signature('s'), variant_level=1),  # noqa
                                           dbus.String(u'Address'): dbus.String(u'00:11:67:D2:AB:EE', variant_level=1),  # noqa
                                           dbus.String(u'Services'): dbus.Array([], signature=dbus.Signature('o'), variant_level=1),  # noqa
                                           dbus.String(u'Blocked'): dbus.Boolean(False, variant_level=1),  # noqa
                                           dbus.String(u'Class'): dbus.UInt32(2360340L, variant_level=1),  # noqa
                                           dbus.String(u'Trusted'): dbus.Boolean(True, variant_level=1),  # noqa
                                           dbus.String(u'Icon'): dbus.String(u'audio-card', variant_level=1)},  # noqa
                                          signature=dbus.Signature('sv'))

            service_xml = (
                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x00010001" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x1200" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
                <uint16 value="0x0001" />
            </sequence>
            <sequence>
                <uuid value="0x0001" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x1200" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0200">
        <uint16 value="0x0103" />
    </attribute>
    <attribute id="0x0201">
        <uint16 value="0x0039" />
    </attribute>
    <attribute id="0x0202">
        <uint16 value="0x13a4" />
    </attribute>
    <attribute id="0x0203">
        <uint16 value="0x0104" />
    </attribute>
    <attribute id="0x0204">
        <boolean value="true" />
    </attribute>
    <attribute id="0x0205">
        <uint16 value="0x0001" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x00010002" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x1108" />
            <uuid value="0x1203" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
            </sequence>
            <sequence>
                <uuid value="0x0003" />
                <uint8 value="0x02" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x1108" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0100">
        <text value="Headset unit" />
    </attribute>
    <attribute id="0x0302">
        <boolean value="true" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x00010003" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x111e" />
            <uuid value="0x1203" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
            </sequence>
            <sequence>
                <uuid value="0x0003" />
                <uint8 value="0x01" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x111e" />
                <uint16 value="0x0105" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0100">
        <text value="Hands-free unit" />
    </attribute>
    <attribute id="0x0311">
        <uint16 value="0x001f" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x00010006" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x110e" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
                <uint16 value="0x0017" />
            </sequence>
            <sequence>
                <uuid value="0x0017" />
                <uint16 value="0x0102" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x110e" />
                <uint16 value="0x0103" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0100">
        <text value="AVRCP CT" />
    </attribute>
    <attribute id="0x0102">
        <text value="ISSC" />
    </attribute>
    <attribute id="0x0311">
        <uint16 value="0x0001" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x00010008" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x110b" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
                <uint16 value="0x0019" />
            </sequence>
            <sequence>
                <uuid value="0x0019" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x110d" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0100">
        <text value="Audio SNK" />
    </attribute>
    <attribute id="0x0102">
        <text value="ISSC" />
    </attribute>
    <attribute id="0x0311">
        <uint16 value="0x0003" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x0001000c" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x110c" />
        </sequence>
    </attribute>
    <attribute id="0x0004">
        <sequence>
            <sequence>
                <uuid value="0x0100" />
                <uint16 value="0x0017" />
            </sequence>
            <sequence>
                <uuid value="0x0017" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x110e" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
    <attribute id="0x0100">
        <text value="AVRCP TG" />
    </attribute>
    <attribute id="0x0102">
        <text value="ISSC" />
    </attribute>
    <attribute id="0x0311">
        <uint16 value="0x0002" />
    </attribute>
</record>
""",

                """<?xml version="1.0" encoding="UTF-8" ?>
<record>
    <attribute id="0x0000">
        <uint32 value="0x0001000f" />
    </attribute>
    <attribute id="0x0001">
        <sequence>
            <uuid value="0x112e" />
        </sequence>
    </attribute>
    <attribute id="0x0006">
        <sequence>
            <uint16 value="0x656e" />
            <uint16 value="0x006a" />
            <uint16 value="0x0100" />
        </sequence>
    </attribute>
    <attribute id="0x0009">
        <sequence>
            <sequence>
                <uuid value="0x1130" />
                <uint16 value="0x0100" />
            </sequence>
        </sequence>
    </attribute>
</record>
"""
                )

            self._services = dbus.Dictionary({dbus.UInt32(65537L): dbus.String(service_xml[0]),  # noqa
                                              dbus.UInt32(65538L): dbus.String(service_xml[1]),  # noqa
                                              dbus.UInt32(65539L): dbus.String(service_xml[2]),  # noqa
                                              dbus.UInt32(65542L): dbus.String(service_xml[3]),  # noqa
                                              dbus.UInt32(65544L): dbus.String(service_xml[4]),  # noqa
                                              dbus.UInt32(65548L): dbus.String(service_xml[5]),  # noqa
                                              dbus.UInt32(65551L): dbus.String(service_xml[6]),  # noqa
                                              }, signature=dbus.Signature('us'))  # noqa

        elif (self.addr == 'org.bluez.Manager'):
            self._props = {u'Adapters': ['/org/bluez/985/hci0']}
        elif (self.addr == 'org.bluez.AudioSink'):
            self._props = {u'Connected': False}
        elif (self.addr == 'org.bluez.Control'):
            self._props = {u'Connected': False}
        elif (self.addr == 'org.bluez.Media'):
            pass
        elif (self.addr == 'org.freedesktop.DBus.Introspectable'):
            self._introspect = \
                """
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
    <interface name="org.freedesktop.DBus.Introspectable">
        <method name="Introspect">
            <arg type="s" direction="out"/>
        </method>
    </interface>
    <interface name="org.bluez.Device">
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <method name="SetProperty">
            <arg type="s" direction="in"/>
            <arg type="v" direction="in"/>
        </method>
        <method name="DiscoverServices">
            <arg type="s" direction="in"/>
            <arg type="a{us}" direction="out"/>
        </method>
        <method name="CancelDiscovery"/>
        <method name="Disconnect"/>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
        <signal name="DisconnectRequested"/>
    </interface>
    <interface name="org.bluez.Serial">
        <method name="Connect">
            <arg type="s" direction="in"/>
            <arg type="s" direction="out"/>
        </method>
        <method name="ConnectFD">
            <arg type="s" direction="in"/>
            <arg type="h" direction="out"/>
        </method>
        <method name="Disconnect">
            <arg type="s" direction="in"/>
        </method>
    </interface>
    <interface name="org.bluez.Input">
        <method name="Connect"/>
        <method name="Disconnect"/>
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
    </interface>
    <interface name="org.bluez.Audio">
        <method name="Connect"/>
        <method name="Disconnect"/>
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
    </interface>
    <interface name="org.bluez.Headset">
        <method name="Connect"/>
        <method name="Disconnect"/>
        <method name="IsConnected">
            <arg type="b" direction="out"/>
        </method>
        <method name="IndicateCall"/>
        <method name="CancelCall"/>
        <method name="Play"/>
        <method name="Stop"/>
        <method name="IsPlaying">
            <arg type="b" direction="out"/>
        </method>
        <method name="GetSpeakerGain">
            <arg type="q" direction="out"/>
        </method>
        <method name="GetMicrophoneGain">
            <arg type="q" direction="out"/>
        </method>
        <method name="SetSpeakerGain">
            <arg type="q" direction="in"/>
        </method>
        <method name="SetMicrophoneGain">
            <arg type="q" direction="in"/>
        </method>
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <method name="SetProperty">
            <arg type="s" direction="in"/>
            <arg type="v" direction="in"/>
        </method>
        <signal name="Connected"/>
        <signal name="Disconnected"/>
        <signal name="AnswerRequested"/>
        <signal name="Stopped"/>
        <signal name="Playing"/>
        <signal name="SpeakerGainChanged">
            <arg type="q"/>
        </signal>
        <signal name="MicrophoneGainChanged">
            <arg type="q"/>
        </signal>
        <signal name="CallTerminated"/>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
    </interface>
    <interface name="org.bluez.AudioSink">
        <method name="Connect"/>
        <method name="Disconnect"/>
        <method name="IsConnected">
            <arg type="b" direction="out"/>
        </method>
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <signal name="Connected"/>
        <signal name="Disconnected"/>
        <signal name="Playing"/>
        <signal name="Stopped"/>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
    </interface>
    <interface name="org.bluez.Control">
        <method name="IsConnected">
            <arg type="b" direction="out"/>
        </method>
        <method name="GetProperties">
            <arg type="a{sv}" direction="out"/>
        </method>
        <method name="VolumeUp"/>
        <method name="VolumeDown"/>
        <signal name="Connected"/>
        <signal name="Disconnected"/>
        <signal name="PropertyChanged">
            <arg type="s"/>
            <arg type="v"/>
        </signal>
    </interface>
</node>
"""

    def Introspect(self, addr):
        return self._introspect[1:]  # Strip leading \n

    def StartDiscovery(self):
        self._props[dbus.String(u'Discovering')] = \
            dbus.Boolean(True, variant_level=1)

    def StopDiscovery(self):
        self._props[dbus.String(u'Discovering')] = \
            dbus.Boolean(False, variant_level=1)

    def ListAdapters(self):
        return ['/org/bluez/985/hci0']

    def FindAdapter(self, *args):
        return '/org/bluez/985/hci0'

    def DefaultAdapter(self, *args):
        return '/org/bluez/985/hci0'

    def SetProperty(self, name, value):
        self._props[name] = value

    def GetProperties(self):
        return self._props

    def FindDevice(self, *args):
        return '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'

    def ListDevices(self, *args):
        return ['/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE']

    def CreatePairedDevice(self, dev_id, path, caps,
                           reply_handler, error_handler):
        self._cb_notify_dev_id = dev_id
        self._cb_notify_device = reply_handler
        self._cb_notify_error = error_handler

    def RemoveDevice(self, dev_obj):
        pass

    def RegisterAgent(self, path, caps):
        pass

    def UnregisterAgent(self, path):
        pass

    def Connect(self):
        self._props['Connected'] = True

    def IsConnected(self):
        return self._props['Connected']

    def Disconnect(self):
        self._props['Connected'] = False

    def VolumeUp(self):
        pass

    def VolumeDown(self):
        pass

    def DiscoverServices(self, pattern):
        return self._services

    def RegisterEndpoint(self, path, properties):
        pass

    def RegisterPlayer(self, path, properties):
        pass

    def UnregisterEndpoint(self, path):
        pass

    def UnregisterPlayer(self, path):
        pass

    def Acquire(self, access_mode):
        pass

    def Release(self, access_mode):
        pass

    def _test_notify_device_created_ok(self):
        self._cb_notify_device('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')  # noqa

    def _test_notify_device_created_error(self):
        self._cb_notify_error('Unable to create device: ' +
                              self._cb_notify_dev_id)


class BTDbusTypeTranslation(unittest.TestCase):

    def test_all(self):

        val = [1, 2, 3, 4, 5]
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.Array, val),  # noqa
                         dbus.Array(val))
        val = -1
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.Int32, val),  # noqa
                         dbus.Int32(val))
        val = 1
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.UInt32, val),  # noqa
                         dbus.UInt32(val))
        val = 'Test'
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.String, val),  # noqa
                         dbus.String(val))
        val = {'Hello': 1}
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.Dictionary,  # noqa
                                                           val),
                         dbus.Dictionary(val))
        val = 'True'
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.Boolean, val),  # noqa
                         dbus.Boolean(True))
        val = 'False'
        self.assertEqual(bt_manager.interface.translate_to_dbus_type(dbus.Boolean, val),  # noqa
                         dbus.Boolean(False))


class BTManagerTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')
        self.mock_system_bus = mock_system_bus

    def test_manager_adapters(self):
        manager = bt_manager.BTManager()
        print repr(manager)
        print '========================================================='
        print manager
        print '========================================================='
        print 'Adapters:', manager.list_adapters()
        print 'Default:', manager.default_adapter()


class BTAdapterTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')
        self.mock_system_bus = mock_system_bus

    def test_get_adapter_properties(self):
        adapter = bt_manager.BTAdapter()
        print repr(adapter)
        print '========================================================='
        print adapter
        print '========================================================='
        cod = bt_manager.BTCoD(adapter.Class)
        print 'Adapter Class:', hex(adapter.Class)
        print 'Major Service Class:', str(cod.major_service_class)
        print 'Major Device Class:', str(cod.major_device_class)
        print 'Minor Device Class:', str(cod.minor_device_class)
        print '========================================================='
        uuids = adapter.UUIDs
        for i in uuids:
            uuid = bt_manager.BTUUID(i)
            print bt_manager.SERVICES.get(uuid.uuid16, 'Unknown')
        print '========================================================='

    def test_get_adapter_by_dev_id(self):
        adapter = bt_manager.BTAdapter()
        adapter_by_dev_id = bt_manager.BTAdapter(adapter.Address)
        self.assertEqual(adapter_by_dev_id.Address, adapter.Address)

    def test_set_adapter_property(self):
        adapter = bt_manager.BTAdapter()
        new_name = 'NewAdapterName-0'
        adapter.Name = new_name
        self.assertEqual(adapter.Name, new_name)

    def test_adapter_list_devices(self):
        adapter = bt_manager.BTAdapter()
        print adapter.list_devices()
        print '========================================================='

    def test_adapter_discovery(self):
        adapter = bt_manager.BTAdapter()
        adapter.start_discovery()
        self.assertTrue(adapter.Discovering)
        adapter.stop_discovery()
        self.assertFalse(adapter.Discovering)

    def test_adapter_signal_device_found(self):
        name = '11:22:33:44:55:66'
        properties = 'Test Properties'
        signal = bt_manager.BTAdapter.SIGNAL_DEVICE_FOUND

        user = mock.MagicMock()
        adapter = bt_manager.BTAdapter()
        adapter.add_signal_receiver(user.callback_fn, signal, self)
        self.mock_system_bus.add_signal_receiver.assert_called()
        cb = self.mock_system_bus.add_signal_receiver.call_args_list[0][0][0]
        cb(name, properties)
        user.callback_fn.assert_called_with(signal, self, name, properties)
        adapter.remove_signal_receiver(signal)
        self.mock_system_bus.remove_signal_receiver.assert_called()

    def test_adapter_signal_property_changed(self):
        name = 'Property'
        value = 'New Value'
        signal = bt_manager.BTAdapter.SIGNAL_PROPERTY_CHANGED

        user = mock.MagicMock()
        adapter = bt_manager.BTAdapter()
        adapter.add_signal_receiver(user.callback_fn, signal, self)
        self.mock_system_bus.add_signal_receiver.assert_called()
        cb = self.mock_system_bus.add_signal_receiver.call_args_list[0][0][0]
        cb(name, value)
        user.callback_fn.assert_called_with(signal, self, name, value)
        adapter.remove_signal_receiver(signal)
        self.mock_system_bus.remove_signal_receiver.assert_called()

    def test_adapter_signal_name_exception(self):
        adapter = bt_manager.BTAdapter()
        try:
            exception_caught = False
            adapter.add_signal_receiver(None, 'Invalid Signal Name', None)
        except bt_manager.BTSignalNameNotRecognisedException:
            exception_caught = True
        self.assertTrue(exception_caught)

    def test_adapter_find_device(self):
        adapter = bt_manager.BTAdapter()
        dev = adapter.find_device('00:11:67:D2:AB:EE')
        self.assertEqual(dev,
                         dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'))  # noqa

    def test_adapter_create_device(self):

        user = mock.MagicMock()
        adapter = bt_manager.BTAdapter()
        dev_id = '00:11:67:D2:AB:EE'
        path = '/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'
        caps = None

        adapter.create_paired_device(dev_id, path,
                                     caps, user.cb_notify_device,
                                     user.cb_notify_error)
        adapter._interface._test_notify_device_created_ok()
        user.cb_notify_device.assert_called()
        adapter._interface._test_notify_device_created_error()
        user.cb_notify_error.assert_called()

    def test_adapter_remove_device(self):
        adapter = bt_manager.BTAdapter()
        adapter.remove_device('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')

    def test_adapter_register_agent(self):
        adapter = bt_manager.BTAdapter()
        adapter.register_agent('/test/agent', 'DisplayYesNo')
        adapter.unregister_agent('/test/agent')


class BTDeviceTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

    def test_get_device_properties(self):
        adapter = bt_manager.BTAdapter()
        adapter_id = adapter.Address
        devices = adapter.list_devices()
        dev_path = devices[0]
        device = bt_manager.BTDevice(dev_path=dev_path)
        print 'Vendor Name:', bt_manager.VENDORS.get(device.Vendor, 'Unknown')
        print device
        print '========================================================='
        cod = bt_manager.BTCoD(device.Class)
        print 'Device Class:', hex(device.Class)
        print 'Major Service Class:', str(cod.major_service_class)
        print 'Major Device Class:', str(cod.major_device_class)
        print 'Minor Device Class:', str(cod.minor_device_class)
        print '========================================================='
        print repr(cod)
        print '========================================================='
        print cod
        print '========================================================='
        uuids = device.UUIDs
        for i in uuids:
            uuid = bt_manager.BTUUID(i)
            print bt_manager.SERVICES.get(uuid.uuid16, 'Unknown')
        print '========================================================='
        device = bt_manager.BTDevice(adapter_id=adapter_id,
                                     dev_path=dev_path)
        print device
        print '========================================================='
        device = bt_manager.BTDevice(dev_id='00:11:67:D2:AB:EE')
        print device
        print '========================================================='
        print repr(device)
        print '========================================================='

    def test_set_device_property(self):
        devices = bt_manager.BTAdapter().list_devices()
        dev_path = devices[0]
        device = bt_manager.BTDevice(dev_path=dev_path)
        device.Trusted = False
        self.assertEqual(device.Trusted, False)

    def test_discover_services(self):
        device = bt_manager.BTDevice(dev_id='00:11:67:D2:AB:EE')
        services = device.discover_services()
        print '========================================================='
        for rec in services.keys():
            print bt_manager.BTDiscoveryInfo(services[rec])
        print '========================================================='


class BTAudioSinkTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

    def test_audio_sink_properties(self):
        sink = bt_manager.BTAudioSink(dev_id='00:11:67:D2:AB:EE')
        print sink
        print '========================================================='
        print repr(sink)
        print '========================================================='

    def test_audio_sink_connectivity(self):
        sink = bt_manager.BTAudioSink(dev_id='00:11:67:D2:AB:EE')
        sink.connect()
        self.assertTrue(sink.is_connected())
        sink.disconnect()
        self.assertFalse(sink.is_connected())


class BTControlTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

    def test_control_properties(self):
        ctrl = bt_manager.BTControl(dev_id='00:11:67:D2:AB:EE')
        print ctrl
        print '========================================================='
        print repr(ctrl)
        print '========================================================='

    def test_control_volume(self):
        ctrl = bt_manager.BTControl(dev_id='00:11:67:D2:AB:EE')
        ctrl.volume_up()
        ctrl.volume_down()


class BTMediaTest(unittest.TestCase):

    def setUp(self):
        patcher = mock.patch('dbus.Interface', MockDBusInterface)
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('dbus.SystemBus')
        patched_system_bus = patcher.start()
        self.addCleanup(patcher.stop)
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

    def test_media_endpoint(self):
        media = bt_manager.BTMedia()
        path = '/test/sbcaudiosink'
        endpoint = bt_manager.SBCAudioSink(path=path)
        media.register_endpoint(path, endpoint.get_properties())
        media.unregister_endpoint(path)


class BTAgentTest(unittest.TestCase):

    @mock.patch('dbus.SystemBus')
    def test_agent_with_defaults(self, patched_system_bus):
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')
        agent = bt_manager.BTAgent()

        obj = dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')
        uuid = dbus.String(u'00001108-0000-1000-8000-00805f9b34fb')
        pin_code = dbus.String('0000')
        pass_key = dbus.UInt32(0)
        mode = 'Mode'

        self.assertEqual(agent.Release(), None)
        self.assertEqual(agent.Authorize(obj, uuid), None)
        self.assertEqual(agent.RequestPinCode(obj), pin_code)
        self.assertEqual(agent.RequestPasskey(obj), pass_key)
        self.assertEqual(agent.DisplayPasskey(obj, pass_key), None)
        self.assertEqual(agent.RequestConfirmation(obj, pass_key), None)
        self.assertEqual(agent.ConfirmModeChange(mode), None)
        self.assertEqual(agent.Cancel(), None)

    @mock.patch('dbus.SystemBus')
    def test_agent_with_callbacks(self, patched_system_bus):
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')
        user = mock.MagicMock()
        agent = bt_manager.BTAgent(cb_notify_on_release=user.cb_notify_on_release,  # noqa
                                   cb_notify_on_authorize=user.cb_notify_on_authorize,  # noqa
                                   cb_notify_on_request_pin_code=user.cb_notify_on_request_pin_code,  # noqa
                                   cb_notify_on_request_pass_key=user.cb_notify_on_request_pass_key,  # noqa
                                   cb_notify_on_display_pass_key=user.cb_notify_on_display_pass_key,  # noqa
                                   cb_notify_on_request_confirmation=user.cb_notify_on_request_confirmation,  # noqa
                                   cb_notify_on_confirm_mode_change=user.cb_notify_on_confirm_mode_change,  # noqa
                                   cb_notify_on_cancel=user.cb_notify_on_cancel)  # noqa

        obj = dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')
        uuid = dbus.String(u'00001108-0000-1000-8000-00805f9b34fb')
        pin_code = dbus.String('0000')
        pass_key = dbus.UInt32(0x12345678L)
        mode = 'Mode'

        self.assertEqual(agent.Release(), None)
        user.cb_notify_on_release.assert_once_called_with()

        user.cb_notify_on_authorize.return_value = True
        self.assertEqual(agent.Authorize(obj, uuid), None)
        user.cb_notify_on_authorize.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_AUTHORIZE,  # noqa
                                                            obj, uuid)

        user.reset_mock()
        user.cb_notify_on_authorize.return_value = False
        try:
            exception_raised = False
            agent.Authorize(obj, uuid)
        except bt_manager.BTRejectedException:
            exception_raised = True
        user.cb_notify_on_authorize.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_AUTHORIZE,  # noqa
                                                            obj, uuid)
        self.assertTrue(exception_raised)

        user.cb_notify_on_request_pin_code.return_value = pin_code
        self.assertEqual(agent.RequestPinCode(obj), pin_code)
        user.cb_notify_on_request_pin_code.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_PIN_CODE,  # noqa
                                                                   obj)

        user.reset_mock()
        user.cb_notify_on_request_pin_code.return_value = None
        try:
            exception_raised = False
            agent.RequestPinCode(obj)
        except bt_manager.BTRejectedException:
            exception_raised = True
        user.cb_notify_on_request_pin_code.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_PIN_CODE,  # noqa
                                                                   obj)
        self.assertTrue(exception_raised)

        user.cb_notify_on_request_pass_key.return_value = pass_key
        self.assertEqual(agent.RequestPasskey(obj), pass_key)
        user.cb_notify_on_request_pass_key.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_PASS_KEY,  # noqa
                                                                   obj)

        user.reset_mock()
        user.cb_notify_on_request_pass_key.return_value = None
        try:
            exception_raised = False
            agent.RequestPasskey(obj)
        except bt_manager.BTRejectedException:
            exception_raised = True
        user.cb_notify_on_request_pass_key.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_PASS_KEY,  # noqa
                                                                   obj)
        self.assertTrue(exception_raised)

        self.assertEqual(agent.DisplayPasskey(obj, pass_key), None)
        user.cb_notify_on_display_pass_key.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_DISPLAY_PASS_KEY,  # noqa
                                                                   obj,
                                                                   pass_key)

        user.cb_notify_on_request_confirmation.return_value = True
        self.assertEqual(agent.RequestConfirmation(obj, pass_key), None)
        user.cb_notify_on_request_confirmation.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_CONFIRMATION,  # noqa
                                                                       obj,
                                                                       pass_key)  # noqa

        user.reset_mock()
        user.cb_notify_on_request_confirmation.return_value = False
        try:
            exception_raised = False
            agent.RequestConfirmation(obj, pass_key)
        except bt_manager.BTRejectedException:
            exception_raised = True
        user.cb_notify_on_request_confirmation.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_REQUEST_CONFIRMATION,  # noqa
                                                                       obj,
                                                                       pass_key)  # noqa
        self.assertTrue(exception_raised)

        user.cb_notify_on_confirm_mode_change.return_value = True
        self.assertEqual(agent.ConfirmModeChange(mode), None)
        user.cb_notify_on_confirm_mode_change.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_CONFIRM_MODE_CHANGE,  # noqa
                                                                      mode)

        user.reset_mock()
        user.cb_notify_on_confirm_mode_change.return_value = False
        try:
            exception_raised = False
            agent.ConfirmModeChange(mode)
        except bt_manager.BTRejectedException:
            exception_raised = True
        user.cb_notify_on_confirm_mode_change.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_CONFIRM_MODE_CHANGE,  # noqa
                                                                      mode)
        self.assertTrue(exception_raised)

        self.assertEqual(agent.Cancel(), None)
        user.cb_notify_on_cancel.assert_called_once_with(bt_manager.BTAgent.NOTIFY_ON_CANCEL)  # noqa

    @mock.patch('dbus.SystemBus')
    def test_agent_corner_cases(self, patched_system_bus):
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')
        agent = bt_manager.BTAgent(default_pin_code=None,
                                   default_pass_key=None)

        obj = dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE')

        try:
            exception_raised = False
            agent.RequestPinCode(obj)
        except bt_manager.BTRejectedException:
            exception_raised = True
        self.assertTrue(exception_raised)

        try:
            exception_raised = False
            agent.RequestPasskey(obj)
        except bt_manager.BTRejectedException:
            exception_raised = True
        self.assertTrue(exception_raised)


class SBCAudioTest(unittest.TestCase):

    @mock.patch('dbus.SystemBus')
    def test_sbc_caps_conversion(self, patched_system_bus):

        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

        media = bt_manager.SBCAudioCodec(uuid='uuid', path='/endpoint/test')
        config = bt_manager.SBCCodecConfig(bt_manager.SBCChannelMode.ALL,
                                           bt_manager.SBCSamplingFrequency.ALL,
                                           bt_manager.SBCAllocationMethod.ALL,
                                           bt_manager.SBCSubbands.ALL,
                                           bt_manager.SBCBlocks.ALL,
                                           2,
                                           64)
        dbus_config = media._make_config(config)
        self.assertEqual(dbus_config, dbus.Array([dbus.Byte(0xFF),
                                                  dbus.Byte(0xFF),
                                                  dbus.Byte(2),
                                                  dbus.Byte(64)]))
        self.assertEqual(media._parse_config(dbus_config), config)

    def test_sbc_default_bitpool(self):

        frequency = bt_manager.SBCSamplingFrequency.FREQ_16KHZ
        channel_mode = 0   # Don't care
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 53)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_32KHZ
        channel_mode = 0   # Don't care
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 53)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_MONO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 31)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_DUAL
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 31)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_JOINT_STEREO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 53)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_STEREO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 53)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_48KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_STEREO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 51)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_48KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_JOINT_STEREO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 51)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_48KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_MONO
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 29)

        frequency = bt_manager.SBCSamplingFrequency.FREQ_48KHZ
        channel_mode = bt_manager.SBCChannelMode.CHANNEL_MODE_DUAL
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 29)

        frequency = -1     # Out of range
        channel_mode = -1  # Out of range
        bitpool = bt_manager.SBCAudioCodec._default_bitpool(frequency,
                                                            channel_mode)
        self.assertEqual(bitpool, 53)

    @mock.patch('dbus.SystemBus')
    def test_sbc_caps_negotiation(self, patched_system_bus):
        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

        media = bt_manager.SBCAudioSource()
        caps = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.ALL,
                                         bt_manager.codecs.SBCSamplingFrequency.ALL,  # noqa
                                         bt_manager.codecs.SBCAllocationMethod.ALL,  # noqa
                                         bt_manager.codecs.SBCSubbands.ALL,
                                         bt_manager.codecs.SBCBlocks.ALL,
                                         2,
                                         64)
        expected = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_JOINT_STEREO,  # noqa
                                             bt_manager.codecs.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                             bt_manager.codecs.SBCAllocationMethod.LOUDNESS,  # noqa
                                             bt_manager.codecs.SBCSubbands.SUBBANDS_8,  # noqa
                                             bt_manager.codecs.SBCBlocks.BLOCKS_16,  # noqa
                                             2,
                                             53)
        dbus_caps = media._make_config(caps)
        expected_dbus = media._make_config(expected)
        actual_dbus = media.SelectConfiguration(dbus_caps)
        self.assertEqual(actual_dbus, expected_dbus)

        caps = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_MONO,  # noqa
                                         bt_manager.codecs.SBCSamplingFrequency.FREQ_48KHZ,  # noqa
                                         bt_manager.codecs.SBCAllocationMethod.SNR,  # noqa
                                         bt_manager.codecs.SBCSubbands.SUBBANDS_4,  # noqa
                                         bt_manager.codecs.SBCBlocks.BLOCKS_12,
                                         2,
                                         64)
        expected = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_MONO,  # noqa
                                             bt_manager.codecs.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                             bt_manager.codecs.SBCAllocationMethod.SNR,  # noqa
                                             bt_manager.codecs.SBCSubbands.SUBBANDS_4,  # noqa
                                             bt_manager.codecs.SBCBlocks.BLOCKS_12,  # noqa
                                             2,
                                             31)
        dbus_caps = media._make_config(caps)
        expected_dbus = media._make_config(expected)
        actual_dbus = media.SelectConfiguration(dbus_caps)
        self.assertEqual(actual_dbus, expected_dbus)

        caps = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_DUAL,  # noqa
                                         bt_manager.codecs.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                         bt_manager.codecs.SBCAllocationMethod.LOUDNESS,  # noqa
                                         bt_manager.codecs.SBCSubbands.SUBBANDS_8,  # noqa
                                         bt_manager.codecs.SBCBlocks.BLOCKS_8,
                                         2,
                                         64)
        expected = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_DUAL,  # noqa
                                             bt_manager.codecs.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                             bt_manager.codecs.SBCAllocationMethod.LOUDNESS,  # noqa
                                             bt_manager.codecs.SBCSubbands.SUBBANDS_8,  # noqa
                                             bt_manager.codecs.SBCBlocks.BLOCKS_8,  # noqa
                                             2,
                                             31)
        dbus_caps = media._make_config(caps)
        expected_dbus = media._make_config(expected)
        actual_dbus = media.SelectConfiguration(dbus_caps)
        self.assertEqual(actual_dbus, expected_dbus)

        caps = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_STEREO,  # noqa
                                         bt_manager.codecs.SBCSamplingFrequency.FREQ_32KHZ,  # noqa
                                         bt_manager.codecs.SBCAllocationMethod.ALL,  # noqa
                                         bt_manager.codecs.SBCSubbands.SUBBANDS_8,  # noqa
                                         bt_manager.codecs.SBCBlocks.BLOCKS_4,
                                         2,
                                         64)
        expected = bt_manager.SBCCodecConfig(bt_manager.codecs.SBCChannelMode.CHANNEL_MODE_STEREO,  # noqa
                                             bt_manager.codecs.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                             bt_manager.codecs.SBCAllocationMethod.LOUDNESS,  # noqa
                                             bt_manager.codecs.SBCSubbands.SUBBANDS_8,  # noqa
                                             bt_manager.codecs.SBCBlocks.BLOCKS_4,  # noqa
                                             2,
                                             53)
        dbus_caps = media._make_config(caps)
        expected_dbus = media._make_config(expected)
        actual_dbus = media.SelectConfiguration(dbus_caps)
        self.assertEqual(actual_dbus, expected_dbus)

    @mock.patch('os.close')
    @mock.patch('dbus.SystemBus')
    @mock.patch('bt_manager.audio.BTAudioSink')
    @mock.patch('bt_manager.audio.BTMediaTransport')
    def test_sbc_audio_source(self, patched_transport, patched_audio,
                              patched_system_bus, mock_close):

        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

        mock_audio = mock.MagicMock()
        patched_audio.return_value = mock_audio
        patched_audio.SIGNAL_PROPERTY_CHANGED = 'PropertyChanged'
        mock_audio.State = 'disconnected'

        mock_transport = mock.MagicMock()
        patched_transport.return_value = mock_transport

        media = bt_manager.SBCAudioSource()
        caps = bt_manager.SBCCodecConfig(bt_manager.SBCChannelMode.ALL,
                                         bt_manager.SBCSamplingFrequency.ALL,
                                         bt_manager.SBCAllocationMethod.ALL,
                                         bt_manager.SBCSubbands.ALL,
                                         bt_manager.SBCBlocks.ALL,
                                         2,
                                         64)
        dbus_caps = media._make_config(caps)
        transport = dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE/fd0')  # noqa
        config = bt_manager.SBCCodecConfig(bt_manager.SBCChannelMode.CHANNEL_MODE_JOINT_STEREO,  # noqa
                                           bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                           bt_manager.SBCAllocationMethod.LOUDNESS,  # noqa
                                           bt_manager.SBCSubbands.SUBBANDS_8,
                                           bt_manager.SBCBlocks.BLOCKS_16,
                                           2,
                                           53)
        dbus_config = dbus.Dictionary({'Device': dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'),  # noqa
                                       'Configuration': media._make_config(config)})  # noqa
        media.SelectConfiguration(dbus_caps)
        media.SetConfiguration(transport, dbus_config)
        mock_audio.add_signal_receiver.assert_called_once_with(media._property_change_event_handler,  # noqa
                                                               bt_manager.BTAudioSink.SIGNAL_PROPERTY_CHANGED,  # noqa
                                                               transport)

        fd_value = os.open('/dev/null', os.O_WRONLY)
        read_mtu = 503
        write_mtu = 503
        fd = mock.MagicMock()
        fd.take.return_value = fd_value
        mock_transport.acquire.return_value = (fd, write_mtu, read_mtu)

        mock_audio.State = 'connecting'
        media._property_change_event_handler('State',
                                             transport, mock_audio.State)
        mock_audio.State = 'connected'
        media._property_change_event_handler('State',
                                             transport, mock_audio.State)
        mock_transport.acquire.assert_called_once_with('w')
        mock_audio.State = 'disconnected'

        data = [b'\x00'] * 512
        media.write_transport(data)

        try:
            exception_caught = False
            media.read_transport()
        except bt_manager.BTIncompatibleTransportAccessType:
            exception_caught = True
        self.assertTrue(exception_caught)

        media._property_change_event_handler('State', transport,
                                             mock_audio.State)
        mock_transport.release.assert_called_once_with('w')
        mock_close.assert_called_with(fd_value)

        media.ClearConfiguration()
        media.Release()

    @mock.patch('os.close')
    @mock.patch('dbus.SystemBus')
    @mock.patch('bt_manager.audio.BTAudioSource')
    @mock.patch('bt_manager.audio.BTMediaTransport')
    def test_sbc_audio_sink(self, patched_transport, patched_audio,
                            patched_system_bus, mock_close):

        mock_system_bus = mock.MagicMock()
        patched_system_bus.return_value = mock_system_bus
        mock_system_bus.get_object.return_value = dbus.ObjectPath('/org/bluez')

        mock_audio = mock.MagicMock()
        patched_audio.return_value = mock_audio
        patched_audio.SIGNAL_PROPERTY_CHANGED = 'PropertyChanged'
        mock_audio.State = 'disconnected'

        mock_transport = mock.MagicMock()
        patched_transport.return_value = mock_transport

        media = bt_manager.SBCAudioSink()
        caps = bt_manager.SBCCodecConfig(bt_manager.SBCChannelMode.ALL,
                                         bt_manager.SBCSamplingFrequency.ALL,
                                         bt_manager.SBCAllocationMethod.ALL,
                                         bt_manager.SBCSubbands.ALL,
                                         bt_manager.SBCBlocks.ALL,
                                         2,
                                         64)
        dbus_caps = media._make_config(caps)
        transport = dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE/fd0')  # noqa
        config = bt_manager.SBCCodecConfig(bt_manager.SBCChannelMode.CHANNEL_MODE_JOINT_STEREO,  # noqa
                                           bt_manager.SBCSamplingFrequency.FREQ_44_1KHZ,  # noqa
                                           bt_manager.SBCAllocationMethod.LOUDNESS,  # noqa
                                           bt_manager.SBCSubbands.SUBBANDS_8,
                                           bt_manager.SBCBlocks.BLOCKS_16,
                                           2,
                                           53)
        dbus_config = dbus.Dictionary({'Device': dbus.ObjectPath('/org/bluez/985/hci0/dev_00_11_67_D2_AB_EE'),  # noqa
                                       'Configuration': media._make_config(config)})  # noqa
        media.SelectConfiguration(dbus_caps)
        media.SetConfiguration(transport, dbus_config)
        mock_audio.add_signal_receiver.assert_called_once_with(media._property_change_event_handler,  # noqa
                                                               'PropertyChanged',  # noqa
                                                               transport)

        fd_value = os.open('tests/vector.dat', os.O_RDONLY)
        read_mtu = 503
        write_mtu = 503
        fd = mock.MagicMock()
        fd.take.return_value = fd_value
        mock_transport.acquire.return_value = (fd, write_mtu, read_mtu)

        mock_audio.State = 'connected'
        media._property_change_event_handler('State',
                                             transport, mock_audio.State)
        mock_audio.State = 'playing'
        media._property_change_event_handler('State',
                                             transport, mock_audio.State)
        mock_transport.acquire.assert_called_once_with('r')
        mock_audio.State = 'connected'

        data = media.read_transport()
        self.assertEqual(len(data), 512)

        try:
            exception_caught = False
            media.write_transport('dummy data')
        except bt_manager.BTIncompatibleTransportAccessType:
            exception_caught = True
        self.assertTrue(exception_caught)

        media._property_change_event_handler('State', transport,
                                             mock_audio.State)
        mock_transport.release.assert_called_once_with('r')
        mock_close.assert_called_with(fd_value)

        media.ClearConfiguration()
        media.Release()



from __future__ import unicode_literals

import unittest

import bt_manager


class UUIDServiceLookup(unittest.TestCase):

    def test_service_lookup(self):
        for i in sorted(bt_manager.SERVICES):
            service = bt_manager.SERVICES[i]
            print service

    def test_attribute_lookup(self):
        for i in sorted(bt_manager.SERVICES):
            service = bt_manager.SERVICES[i]
            uuid16 = service.uuid16
            attribs = bt_manager.ATTRIBUTES.get(uuid16, None)
            print service.name, '['+service.uuid16+']', attribs



from __future__ import unicode_literals

import readline  # noqa
import bt_manager
import sys
import dbus
import dbus.mainloop.glib
import gobject
import signal
from collections import namedtuple


def dump_signal(signal, *args):
    print '\n========================================================='
    print '>>>>>', signal, '<<<<<'
    print args
    print '========================================================='


def agent_event_handler(*args):
    print '\n========================================================='
    print 'Agent event:', args
    return True


def device_created_ok(*args):
    print '\n========================================================='
    print 'New Device Paired:', args


def device_created_error(*args):
    print '\n========================================================='
    print 'Pairing Error:', args


def cmd_help(args):
    global cmd_table

    if (len(args)):
        cmd_list = [args.pop(0)]
    else:
        cmd_list = cmd_table.keys()

    for i in cmd_list:
        print i, cmd_table[i].args, ":", cmd_table[i].desc


def list_adapters(args):
    print '========================================================='
    try:
        print bt_manager.BTManager().list_adapters()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def list_devices(args):
    print '========================================================='
    global adapter

    try:
        print adapter.list_devices()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def adapter_info(args):
    global adapter

    try:
        print '========================================================='
        print adapter
        print '========================================================='
        cod = bt_manager.BTCoD(adapter.Class)
        print 'Vendor Name:', bt_manager.VENDORS.get(adapter.Vendor, 'Unknown')
        print 'Device Class:', hex(adapter.Class)
        print 'Major Service Class:', str(cod.major_service_class)
        print 'Major Device Class:', str(cod.major_device_class)
        print 'Minor Device Class:', str(cod.minor_device_class)
        print '========================================================='
        uuids = adapter.UUIDs
        for i in uuids:
            uuid = bt_manager.BTUUID(i)
            print bt_manager.SERVICES.get(uuid.uuid16, uuid)
        print '========================================================='
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def adapter_get(args):
    if (len(args)):
        name = args.pop(0)
    else:
        name = None

    print '========================================================='
    global adapter

    try:
        print adapter.get_property(name=name)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def adapter_set(args):
    if (len(args) >= 2):
        name = args.pop(0)
        value = args.pop(0)
    else:
        print 'Error: Requires property name and value'
        return

    global adapter

    try:
        adapter.set_property(name, value)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_rm(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Requires device path'
        return

    global adapter

    try:
        adapter.remove_device(dev_path)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_get(args):
    if (len(args)):
        dev_path = args.pop(0)
        if (len(args)):
            name = args.pop(0)
        else:
            name = None
    else:
        print 'Error: Requires device path'
        return

    try:
        device = bt_manager.BTDevice(dev_path=dev_path)
        print '========================================================='
        print device.get_property(name=name)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_set(args):
    if (len(args) >= 2):
        dev_path = args.pop(0)
        name = args.pop(0)
        value = args.pop(0)
    else:
        print 'Error: Requires device path, property name and value'
        return

    try:
        device = bt_manager.BTDevice(dev_path=dev_path)
        device.set_property(name, value)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_info(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        device = bt_manager.BTDevice(dev_path=dev_path)
        print '========================================================='
        print device
        print '========================================================='
        cod = bt_manager.BTCoD(device.Class)
        print 'Vendor Name:', bt_manager.VENDORS.get(device.Vendor, 'Unknown')
        print 'Device Class:', hex(device.Class)
        print 'Major Service Class:', str(cod.major_service_class)
        print 'Major Device Class:', str(cod.major_device_class)
        print 'Minor Device Class:', str(cod.minor_device_class)
        print '========================================================='
        uuids = device.UUIDs
        for i in uuids:
            uuid = bt_manager.BTUUID(i)
            print bt_manager.SERVICES.get(uuid.uuid16, uuid)
        print '========================================================='
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_disconnect(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        device = bt_manager.BTDevice(dev_path=dev_path)
        device.disconnect()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def device_discovery(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        device = bt_manager.BTDevice(dev_path=dev_path)
        discovery = device.discover_services()
        if (discovery):
            for rec in discovery.keys():
                print '========================================================='  # noqa
                print bt_manager.BTDiscoveryInfo(discovery[rec])
        print '========================================================='
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def discovery_start(args):

    global adapter

    try:
        adapter.start_discovery()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def discovery_stop(args):

    global adapter

    try:
        adapter.stop_discovery()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def agent_start(args):

    global services, adapter
    dev_id = None

    if (len(args)):
        path = args.pop(0)
        if (len(args)):
            dev_id = args.pop(0)
    else:
        print 'Error: Must provide agent path e.g., /test/agent'
        return

    try:
        agent = bt_manager.BTAgent(path=path,
                                   cb_notify_on_release=agent_event_handler,
                                   cb_notify_on_authorize=agent_event_handler,
                                   cb_notify_on_request_confirmation=agent_event_handler,  # noqa
                                   cb_notify_on_confirm_mode_change=agent_event_handler,   # noqa
                                   cb_notify_on_cancel=agent_event_handler)
        services[path] = agent
        caps = 'DisplayYesNo'

        if (dev_id):
            adapter.create_paired_device(dev_id, path, caps,
                                         device_created_ok,
                                         device_created_error)
        else:
            adapter.register_agent(path, caps)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def agent_stop(args):

    global agent, services

    if (len(args)):
        path = args.pop(0)
    else:
        print 'Error: Must provide agent path e.g., /test/agent'
        return

    try:
        adapter.unregister_agent(path)
        services[path].remove_from_connection()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def exit_cleanup(args):
    sys.exit(0)


def sink_info(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        sink = bt_manager.BTAudioSink(dev_path=dev_path)
        print '========================================================='
        print sink
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def sink_connect(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        sink = bt_manager.BTAudioSink(dev_path=dev_path)
        sink.connect()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def sink_disconnect(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        sink = bt_manager.BTAudioSink(dev_path=dev_path)
        sink.disconnect()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def source_info(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        source = bt_manager.BTAudioSource(dev_path=dev_path)
        print '========================================================='
        print source
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def source_connect(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        source = bt_manager.BTAudioSource(dev_path=dev_path)
        source.connect()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def source_disconnect(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        source = bt_manager.BTAudioSource(dev_path=dev_path)
        source.disconnect()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def control_info(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        ctrl = bt_manager.BTControl(dev_path=dev_path)
        print '========================================================='
        print ctrl
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def control_vol_up(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        ctrl = bt_manager.BTControl(dev_path=dev_path)
        ctrl.volume_up()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def control_vol_down(args):
    if (len(args)):
        dev_path = args.pop(0)
    else:
        print 'Error: Must specify device path'
        return

    try:
        ctrl = bt_manager.BTControl(dev_path=dev_path)
        ctrl.volume_down()
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def media_sbc_sink_start(args):

    global services

    if (len(args)):
        path = args.pop(0)
    else:
        print 'Error: Must provide endpoint path e.g., /test/endpoint/sbc0'
        return

    try:
        ep = bt_manager.SBCAudioSink(path=path)
        print '========================================================='
        print repr(ep)
        services[path] = ep
        media = bt_manager.BTMedia()
        media.register_endpoint(path, ep.get_properties())
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def media_decode_handler(args):
    ep = args[0]
    fd = args[1]
    data = ep.read_transport()
    fd.write(data)


def media_encode_handler(args):
    ep = args[0]
    fd = args[1]
    data = fd.read(2560)
    ep.write_transport(data)


def media_decode(args):

    global services

    if (len(args) >= 2):
        path = args.pop(0)
        filename = args.pop(0)
    else:
        print 'Error: Must provide endpoint path e.g., /test/endpoint/sbc0 and audio storage filename'  # noqa
        return

    fd = open(filename, 'wb+')
    ep = services[path]
    ep.register_transport_ready_event(media_decode_handler, (ep, fd))


def media_encode(args):

    global services

    if (len(args) >= 2):
        path = args.pop(0)
        filename = args.pop(0)
    else:
        print 'Error: Must provide endpoint path e.g., /test/endpoint/sbc0 and audio storage filename'  # noqa
        return

    fd = open(filename, 'rb')
    ep = services[path]
    ep.register_transport_ready_event(media_encode_handler, (ep, fd))


def media_sbc_source_start(args):

    global services

    if (len(args)):
        path = args.pop(0)
    else:
        print 'Error: Must provide endpoint path e.g., /test/endpoint/sbc0'
        return

    try:
        ep = bt_manager.SBCAudioSource(path=path)
        print '========================================================='
        print repr(ep)
        services[path] = ep
        media = bt_manager.BTMedia()
        media.register_endpoint(path, ep.get_properties())
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


def media_stop(args):

    global services

    if (len(args)):
        path = args.pop(0)
    else:
        print 'Error: Must provide endpoint path e.g., /test/endpoint/sbc0'
        return

    try:
        ep = services[path]
        ep.unregister_transport_ready_event()
        ep.close_transport()
        ep.remove_from_connection()
        media = bt_manager.BTMedia()
        media.unregister_endpoint(path)
    except dbus.exceptions.DBusException:
        print 'Unable to complete:', sys.exc_info()


CmdEntry = namedtuple('CmdEntry', 'func desc args')
cmd_table = {'help': CmdEntry(cmd_help,
                              'Display a list of commands or get help for a specific command',  # noqa
                              '[command]'),
             'list-adapters': CmdEntry(list_adapters,
                                       'Provide a list of available BT adapters',  # noqa
                                       None),
             'list-devices': CmdEntry(list_devices,
                                      'Display a list of paired BT devices',
                                      None),
             'adapter-info': CmdEntry(adapter_info,
                                      'Display information about default BT adapter',  # noqa
                                      None),
             'adapter-get': CmdEntry(adapter_get,
                                     'Get adapter property by name',
                                     '<property>'),
             'adapter-set': CmdEntry(adapter_set,
                                     'Set adapter property by name, value',
                                     '<property> <value>'),
             'device-rm': CmdEntry(device_rm,
                                   'Remove device from adapter',
                                   '<dev_path>'),
             'device-get': CmdEntry(device_get,
                                     'Get device property by name',
                                     '<dev_path> <property>'),
             'device-set': CmdEntry(device_set,
                                     'Set device property by name, value',
                                     '<dev_path> <property> <value>'),
             'device-info': CmdEntry(device_info,
                                     'Display information about a paired BT device',  # noqa
                                     '<dev_path>'),
             'device-disconnect': CmdEntry(device_disconnect,
                                           'Disconnect a BT device',
                                           '<dev_path>'),
             'device-discovery': CmdEntry(device_discovery,
                                          'Run BT device discovery session',
                                          '<dev_path>'),
             'discovery-start': CmdEntry(discovery_start,
                                         'Start device discovery',
                                         None),
             'discovery-stop': CmdEntry(discovery_stop,
                                        'Stop device discovery',
                                        None),
             'agent-start': CmdEntry(agent_start,
                                     'Start pairing agent',
                                     '<agent_path> [dev_id e.g., 11:22:33:44:55:66]'),  # noqa
             'agent-stop': CmdEntry(agent_stop,
                                   'Stop pairing agent',
                                   '<agent_path>'),
             'sink-info': CmdEntry(sink_info,
                                   'Audio sink properties',
                                   '<dev_path>'),
             'sink-connect': CmdEntry(sink_connect,
                                      'Audio sink connect',
                                      '<dev_path>'),
             'sink-disconnect': CmdEntry(sink_disconnect,
                                         'Audio sink connect',
                                         '<dev_path>'),
             'source-info': CmdEntry(source_info,
                                   'Audio source properties',
                                   '<dev_path>'),
             'source-connect': CmdEntry(source_connect,
                                        'Audio source connect',
                                        '<dev_path>'),
             'source-disconnect': CmdEntry(source_disconnect,
                                           'Audio source connect',
                                           '<dev_path>'),
             'control-info': CmdEntry(control_info,
                                      'Control device info',
                                      '<dev_path>'),
             'control-vol-up': CmdEntry(control_vol_up,
                                        'Control volume up',
                                        '<dev_path>'),
             'control-vol-down': CmdEntry(control_vol_down,
                                          'Control volume down',
                                          '<dev_path>'),
             'media-sbc-sink-start': CmdEntry(media_sbc_sink_start,
                                              'Start media endpoint for SBC audio sink (i.e., connects to a source device)',  # noqa
                                              '<endpoint_path>'),
             'media-decode': CmdEntry(media_decode,
                                      'Start media decode for SBC audio sink',
                                      '<endpoint_path> <audio_filename>'),
             'media-encode': CmdEntry(media_encode,
                                      'Start media encode for SBC audio source',  # noqa
                                      '<endpoint_path> <audio_filename>'),
             'media-sbc-source-start': CmdEntry(media_sbc_source_start,
                                                'Start media endpoint for SBC audio source (i.e., connects to a sink device)',  # noqa
                                                '<endpoint_path>'),
             'media-stop': CmdEntry(media_stop,
                                    'Stop media endpoint',
                                    '<endpoint_path>'),
             'exit': CmdEntry(exit_cleanup,
                              'Cleanup and exit',
                              None),
             }


def invoke_bt_command(text):
    if (not text):
        return
    args = text.split(' ')
    cmd = args.pop(0)
    cmd_entry = cmd_table.get(cmd)
    if (cmd_entry):
        cmd_entry.func(args)
    else:
        print 'Error: Command "%s" was not recognized.' % cmd


def timeout_handler(signum, frame):
    while gobject.MainLoop().get_context().pending():
        gobject.MainLoop().get_context().iteration(False)


dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
gobject.threads_init()
signal.signal(signal.SIGALRM, timeout_handler)
signal.setitimer(signal.ITIMER_REAL, 0.01, 0.01)

try:
    adapter = bt_manager.BTAdapter()
    adapter.add_signal_receiver(dump_signal,
                                bt_manager.BTAdapter.SIGNAL_DEVICE_CREATED,
                                None)
    adapter.add_signal_receiver(dump_signal,
                                bt_manager.BTAdapter.SIGNAL_DEVICE_REMOVED,
                                None)
    adapter.add_signal_receiver(dump_signal,
                                bt_manager.BTAdapter.SIGNAL_DEVICE_DISAPPEARED,
                                None)
    adapter.add_signal_receiver(dump_signal,
                                bt_manager.BTAdapter.SIGNAL_DEVICE_FOUND,
                                None)
    adapter.add_signal_receiver(dump_signal,
                                bt_manager.BTAdapter.SIGNAL_PROPERTY_CHANGED,
                                None)
except dbus.exceptions.DBusException:
    print 'Unable to complete:', sys.exc_info()

services = {}

# Main command processing loop
while True:
    text = raw_input("BT> ")
    invoke_bt_command(text)



from __future__ import unicode_literals

import dbus.service
import gobject
import pprint
import os

from device import BTGenericDevice
from media import GenericEndpoint, BTMediaTransport
from codecs import SBCChannelMode, SBCSamplingFrequency, \
    SBCAllocationMethod, SBCSubbands, SBCBlocks, A2DP_CODECS, \
    SBCCodecConfig, SBCCodec
from serviceuuids import SERVICES
from exceptions import BTIncompatibleTransportAccessType, \
    BTInvalidConfiguration


class BTAudio(BTGenericDevice):
    """
    Wrapper around dbus to encapsulate the org.bluez.Audio
    interface.

    :Properties:

    * **State(str) [readonly]**: Possible values: "disconnected",
        "connecting", "connected" with possible state transitions:

      * "disconnected" -> "connecting"
        Either an incoming or outgoing connection attempt
        ongoing.
      * "connecting" -> "disconnected"
        Connection attempt failed
      * "connecting" -> "connected"
        Successfully connected
      * "connected" -> "disconnected"
        Disconnected from the remote device

    See also: :py:class:`.BTGenericDevice` for setup params.
    """
    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.Audio',
                                 *args, **kwargs)

    def connect(self):
        """
        Connect all supported audio profiles on the device.

        :return:

        .. note:: This may invoke any registered media
            endpoints where media profiles are compatible.
        """
        return self._interface.Connect()

    def disconnect(self):
        """
        Disconnect all audio profiles on the device

        :return:

        .. note:: This may release any registered media
            endpoints where media profiles are compatible.
        """
        return self._interface.Disconnect()


class BTAudioSource(BTAudio):
    """
    Wrapper around dbus to encapsulate the org.bluez.AudioSource
    interface.

    See also: :py:class:`.BTAudio`
    """
    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.AudioSource',
                                 *args, **kwargs)


class BTAudioSink(BTAudio):
    """
    Wrapper around dbus to encapsulate the org.bluez.AudioSink
    interface

    * **Connected(boolean) [readonly]**: Indicates if a stream is
        setup to a A2DP sink on the remote device.
    * **Playing(boolean) [readonly]**: Indicates if a stream is
        active to a A2DP sink on the remote device.

    See also: :py:class:`.BTAudio`
    """

    SIGNAL_CONNECTED = 'Connected'
    """
    :signal Connected(signal_name, user_arg): Sent when a successful
        connection has been made to the remote A2DP Sink
    """
    SIGNAL_DISCONNECTED = 'Disconnected'
    """
    :signal Disconnected(signal_name, user_arg): Sent when the device has
        been disconnected from.
    """
    SIGNAL_PLAYING = 'Playing'
    """
    :signal Playing(signal_name, user_arg): Sent when a stream
        with remote device is started.
    """
    SIGNAL_STOPPED = 'Stopped'
    """
    :signal Stopped(signal_name, user_arg): Sent when a stream with
        remote device is suspended.
    """

    def __init__(self, *args, **kwargs):
        BTGenericDevice.__init__(self, addr='org.bluez.AudioSink',
                                 *args, **kwargs)
        self._register_signal_name(BTAudioSink.SIGNAL_CONNECTED)
        self._register_signal_name(BTAudioSink.SIGNAL_DISCONNECTED)
        self._register_signal_name(BTAudioSink.SIGNAL_PLAYING)
        self._register_signal_name(BTAudioSink.SIGNAL_STOPPED)

    def is_connected(self):
        """
        Returns `True` if a stream is setup to a A2DP sink on
        the remote device, `False` otherwise.

        :return Connected: state of `Connected` attribute
        :rtype: boolean
        """
        return self._interface.IsConnected()


class SBCAudioCodec(GenericEndpoint):
    """
    SBCAudioCodec is an implementation of a media endpoint that
    provides common functionality enabling SBC audio source and
    SBC audio sink media endpoints to be established.

    Since certain procedures are specific to whether or not
    the endpoint is a source or sink, in particular the trigger
    points for when the media transport is acquired/release,
    these parts are left to their respective sub-classes.

    SBCAudioCodec handles the following steps in establishing
    an endpoint:

    * Populates `properties` with the capabilities of the codec.
    * `SelectConfiguration`: computes and returns best SBC codec
        configuration parameters based on device capabilities
    * `SetConfiguration`: a sub-class notifier function is called
    * `ClearConfiguration`: nothing is done
    * `Release`: nothing

    In additional to endpoint establishment, the class also has
    transport read and write functions which will handle the
    required SBC media encoding/decoding and RTP encapsulation.

    The user may also register for `transport ready` events
    which allows transport read and write operations to be
    properly synchronized.

    See also: :py:class:`SBCAudioSink` and :py:class:`SBCAudioSource`
    """
    def __init__(self, uuid, path):
        config = SBCCodecConfig(SBCChannelMode.ALL,
                                SBCSamplingFrequency.ALL,
                                SBCAllocationMethod.ALL,
                                SBCSubbands.ALL,
                                SBCBlocks.ALL,
                                2,
                                64)
        caps = SBCAudioCodec._make_config(config)
        codec = dbus.Byte(A2DP_CODECS['SBC'])
        delayed_reporting = dbus.Boolean(True)
        self.tag = None
        self.properties = dbus.Dictionary({'UUID': uuid,
                                           'Codec': codec,
                                           'DelayReporting': delayed_reporting,
                                           'Capabilities': caps})
        GenericEndpoint.__init__(self, path)

    def _transport_ready_handler(self, fd, cb_condition):
        """
        Wrapper for calling user callback routine to notify
        when transport data is ready to read
        """
        self.user_cb(self.user_arg)
        return True

    def register_transport_ready_event(self, user_cb, user_arg):
        """
        Register for transport ready events.  The `transport ready`
        event is raised via a user callback.  If the endpoint
        is configured as a source, then the user may then
        call :py:meth:`write_transport` in order to send data to
        the associated sink.
        Otherwise, if the endpoint is configured as a sink, then
        the user may call :py:meth:`read_transport` to read
        from the associated source instead.

        :param func user_cb: User defined callback function.  It
            must take one parameter which is the user's callback
            argument.
        :param user_arg: User defined callback argument.
        :return:

        See also: :py:meth:`unregister_transport_ready_event`
        """
        self.user_cb = user_cb
        self.user_arg = user_arg

        if ('r' in self.access_type):
            io_event = gobject.IO_IN
        else:
            io_event = gobject.IO_OUT

        self.tag = gobject.io_add_watch(self.fd, io_event,
                                        self._transport_ready_handler)

    def unregister_transport_ready_event(self):
        """
        Unregister previously registered `transport ready`
        events.

        See also: :py:meth:`register_transport_ready_event`
        """
        if (self.tag):
            gobject.source_remove(self.tag)
            self.tag = None

    def read_transport(self):
        """
        Read data from media transport.
        The returned data payload is SBC decoded and has
        all RTP encapsulation removed.

        :return data: Payload data that has been decoded,
            with RTP encapsulation removed.
        :rtype: array{byte}
        """
        if ('r' not in self.access_type):
            raise BTIncompatibleTransportAccessType
        return self.codec.decode(self.fd, self.read_mtu)

    def write_transport(self, data):
        """
        Write data to media transport.  The data is
        encoded using the SBC codec and RTP encapsulated
        before being written to the transport file
        descriptor.

        :param array{byte} data: Payload data to encode,
            encapsulate and send.
        """
        if ('w' not in self.access_type):
            raise BTIncompatibleTransportAccessType
        return self.codec.encode(self.fd, self.write_mtu, data)

    def close_transport(self):
        """
        Forcibly close previously acquired media transport.

        .. note:: The user should first make sure any transport
            event handlers are unregistered first.
        """
        self._release_media_transport(self.path,
                                      self.access_type)

    def _notify_media_transport_available(self, path, transport):
        """
        Subclass should implement this to trigger setup once
        a new media transport is available.
        """
        pass

    def _acquire_media_transport(self, path, access_type):
        """
        Should be called by subclass when it is ready
        to acquire the media transport file descriptor
        """
        transport = BTMediaTransport(path=path)
        (fd, read_mtu, write_mtu) = transport.acquire(access_type)
        self.fd = fd.take()   # We must do the clean-up later
        self.write_mtu = write_mtu
        self.read_mtu = read_mtu
        self.access_type = access_type
        self.path = path

    def _release_media_transport(self, path, access_type):
        """
        Should be called by subclass when it is finished
        with the media transport file descriptor
        """
        try:
            os.close(self.fd)   # Clean-up previously taken fd
            transport = BTMediaTransport(path=path)
            transport.release(access_type)
        except:
            pass

    @staticmethod
    def _default_bitpool(frequency, channel_mode):
        if (frequency ==
                SBCSamplingFrequency.FREQ_16KHZ or
            frequency ==
                SBCSamplingFrequency.FREQ_32KHZ):
            return 53
        elif (frequency ==
                SBCSamplingFrequency.FREQ_44_1KHZ):
            if (channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_MONO or
                channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_DUAL):
                return 31
            elif (channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_STEREO or
                  channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_JOINT_STEREO):
                return 53
            else:
                # TODO: Invalid channel_mode
                return 53
        elif (frequency == SBCSamplingFrequency.FREQ_48KHZ):
            if (channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_MONO or
                channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_DUAL):
                return 29
            elif (channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_STEREO or
                  channel_mode ==
                    SBCChannelMode.CHANNEL_MODE_JOINT_STEREO):
                return 51
            else:
                # TODO: Invalid channel_mode
                return 51
        else:
            # TODO: Invalid frequency
            return 53

    @staticmethod
    def _make_config(config):
        """Helper to turn SBC codec configuration params into a
        a2dp_sbc_t structure usable by bluez"""
        # The SBC config encoding is taken from a2dp_codecs.h, in particular,
        # the a2dp_sbc_t type is converted into a 4-byte array:
        #   uint8_t channel_mode:4
        #   uint8_t frequency:4
        #   uint8_t allocation_method:2
        #   uint8_t subbands:2
        #   uint8_t block_length:4
        #   uint8_t min_bitpool
        #   uint8_t max_bitpool
        return dbus.Array([dbus.Byte(config.channel_mode |
                                     (config.frequency << 4)),
                           dbus.Byte(config.allocation_method |
                                     (config.subbands << 2) |
                                     (config.block_length << 4)),
                           dbus.Byte(config.min_bitpool),
                           dbus.Byte(config.max_bitpool)])

    @staticmethod
    def _parse_config(config):
        """Helper to turn a2dp_sbc_t structure into a
        more usable set of SBC codec configuration params"""
        frequency = config[0] >> 4
        channel_mode = config[0] & 0xF
        allocation_method = config[1] & 0x03
        subbands = (config[1] >> 2) & 0x03
        block_length = (config[1] >> 4) & 0x0F
        min_bitpool = config[2]
        max_bitpool = config[3]
        return SBCCodecConfig(channel_mode, frequency, allocation_method,
                              subbands, block_length, min_bitpool, max_bitpool)

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="", out_signature="")
    def Release(self):
        print('Release')

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="", out_signature="")
    def ClearConfiguration(self):
        print('ClearConfiguration')

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="ay", out_signature="ay")
    def SelectConfiguration(self, caps):
        print('SelectConfiguration(%s)' % caps)
        our_caps = SBCAudioCodec._parse_config(self.properties['Capabilities'])
        device_caps = SBCAudioCodec._parse_config(caps)
        frequency = SBCSamplingFrequency.FREQ_44_1KHZ

        if ((our_caps.channel_mode & device_caps.channel_mode) &
                SBCChannelMode.CHANNEL_MODE_JOINT_STEREO):
            channel_mode = SBCChannelMode.CHANNEL_MODE_JOINT_STEREO
        elif ((our_caps.channel_mode & device_caps.channel_mode) &
              SBCChannelMode.CHANNEL_MODE_STEREO):
            channel_mode = SBCChannelMode.CHANNEL_MODE_STEREO
        elif ((our_caps.channel_mode & device_caps.channel_mode) &
              SBCChannelMode.CHANNEL_MODE_DUAL):
            channel_mode = SBCChannelMode.CHANNEL_MODE_DUAL
        elif ((our_caps.channel_mode & device_caps.channel_mode) &
              SBCChannelMode.CHANNEL_MODE_MONO):
            channel_mode = SBCChannelMode.CHANNEL_MODE_MONO
        else:
            raise BTInvalidConfiguration

        if ((our_caps.block_length & device_caps.block_length) &
                SBCBlocks.BLOCKS_16):
            block_length = SBCBlocks.BLOCKS_16
        elif ((our_caps.block_length & device_caps.block_length) &
              SBCBlocks.BLOCKS_12):
            block_length = SBCBlocks.BLOCKS_12
        elif ((our_caps.block_length & device_caps.block_length) &
              SBCBlocks.BLOCKS_8):
            block_length = SBCBlocks.BLOCKS_8
        elif ((our_caps.block_length & device_caps.block_length) &
              SBCBlocks.BLOCKS_4):
            block_length = SBCBlocks.BLOCKS_4
        else:
            raise BTInvalidConfiguration

        if ((our_caps.subbands & device_caps.subbands) &
                SBCSubbands.SUBBANDS_8):
            subbands = SBCSubbands.SUBBANDS_8
        elif ((our_caps.subbands & device_caps.subbands) &
              SBCSubbands.SUBBANDS_4):
            subbands = SBCSubbands.SUBBANDS_4
        else:
            raise BTInvalidConfiguration

        if ((our_caps.allocation_method & device_caps.allocation_method) &
                SBCAllocationMethod.LOUDNESS):
            allocation_method = SBCAllocationMethod.LOUDNESS
        elif ((our_caps.allocation_method & device_caps.allocation_method) &
              SBCAllocationMethod.SNR):
            allocation_method = SBCAllocationMethod.SNR
        else:
            raise BTInvalidConfiguration

        min_bitpool = max(our_caps.min_bitpool, device_caps.min_bitpool)
        max_bitpool = min(SBCAudioCodec._default_bitpool(frequency,
                                                         channel_mode),
                          device_caps.max_bitpool)

        selected_config = SBCCodecConfig(channel_mode,
                                         frequency,
                                         allocation_method,
                                         subbands,
                                         block_length,
                                         min_bitpool,
                                         max_bitpool)

        # Create SBC codec based on selected configuration
        self.codec = SBCCodec(selected_config)

        dbus_val = SBCAudioCodec._make_config(selected_config)
        return dbus_val

    @dbus.service.method("org.bluez.MediaEndpoint",
                         in_signature="oay", out_signature="")
    def SetConfiguration(self, transport, config):
        print('SetConfiguration(%s, %s)' % (transport, config))
        self._notify_media_transport_available(config.get('Device'), transport)

    def __repr__(self):
        return pprint.pformat(self.__dict__)


class SBCAudioSink(SBCAudioCodec):
    """
    SBC audio sink media endpoint

    SBCAudioSink implies the BT adapter takes on the role of
    a sink and the external device is the source e.g.,
    iPhone, media player.

    Refer to :py:class:`SBCAudioCodec` for basic overview of
    endpoint steps
    """
    def __init__(self,
                 path='/endpoint/a2dpsink'):
        uuid = dbus.String(SERVICES['AudioSink'].uuid)
        SBCAudioCodec.__init__(self, uuid, path)

    def _property_change_event_handler(self, signal, transport, *args):
        """
        Handler for property change event.  We catch certain state
        transitions in order to trigger media transport
        acquisition/release
        """
        current_state = self.source.State
        if (self.state == 'connected' and current_state == 'playing'):
            self._acquire_media_transport(transport, 'r')
        elif (self.state == 'playing' and current_state == 'connected'):
            self._release_media_transport(transport, 'r')
        self.state = current_state

    def _notify_media_transport_available(self, path, transport):
        """
        Called by the endpoint when a new media transport is
        available
        """
        self.source = BTAudioSource(dev_path=path)
        self.state = self.source.State
        self.source.add_signal_receiver(self._property_change_event_handler,
                                        BTAudioSource.SIGNAL_PROPERTY_CHANGED,  # noqa
                                        transport)


class SBCAudioSource(SBCAudioCodec):
    """
    SBC audio source media endpoint.

    SBCAudioSource implies the adapter takes on the role of
    source and the external device is the sink e.g., speaker.

    Refer to :py:class:`SBCAudioCodec` for basic overview of
    endpoint steps
    """
    def __init__(self,
                 path='/endpoint/a2dpsource'):
        uuid = dbus.String(SERVICES['AudioSource'].uuid)
        SBCAudioCodec.__init__(self, uuid, path)

    def _property_change_event_handler(self, signal, transport, *args):
        """
        Handler for property change event.  We catch certain state
        transitions in order to trigger media transport
        acquisition/release
        """
        current_state = self.sink.State
        if ((self.state == 'disconnected' and current_state == 'connected') or
            (self.state == 'connecting' and
                current_state == 'connected')):
            self._acquire_media_transport(transport, 'w')
        elif (self.state == 'connected' and current_state == 'disconnected'):
            self._release_media_transport(transport, 'w')
        self.state = current_state

    def _notify_media_transport_available(self, path, transport):
        """
        Called by the endpoint when a new media transport is
        available
        """
        self.sink = BTAudioSink(dev_path=path)
        self.state = self.sink.State
        self.sink.add_signal_receiver(self._property_change_event_handler,
                                      BTAudioSource.SIGNAL_PROPERTY_CHANGED,  # noqa
                                      transport)



from __future__ import unicode_literals

import re

from setuptools import setup, find_packages


def get_version(filename):
    content = open(filename).read()
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", content))
    return metadata['version']


setup(
    name='BT-Manager',
    version=get_version('bt_manager/__init__.py'),
    url='https://github.com/liamw9534/bt-manager',
    license='Apache License, Version 2.0',
    author='Liam Wickins',
    author_email='liamw9534@gmail.com',
    description='Python-based Bluetooth Device Management',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests', 'tests.*']),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'setuptools',
        'cffi >= 0.7',
    ],
    setup_requires=['cffi >= 0.7'],
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'mock >= 1.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Communications',
    ],
)



from __future__ import unicode_literals
from collections import namedtuple
from bt_manager import ffi

A2DP_CODECS = {'SBC': 0x00,
               'MPEG12': 0x01,
               'MPEG24': 0x02,
               'ATRAC': 0x03,
               }
"""
Enumeration of codec types supported by A2DP profile
"""

SBCCodecConfig = namedtuple('SBCCodecConfig',
                            'channel_mode frequency allocation_method '
                            'subbands block_length min_bitpool '
                            'max_bitpool')
"""
Named tuple collection of SBC A2DP audio profile properties
"""


class SBCSamplingFrequency:
    """Indicates with which sampling frequency the SBC
    frame has been encoded."""
    FREQ_16KHZ = (1 << 3)
    FREQ_32KHZ = (1 << 2)
    FREQ_44_1KHZ = (1 << 1)
    FREQ_48KHZ = 1
    ALL = 0xF


class SBCBlocks:
    """The block size with which the stream has been encoded"""
    BLOCKS_4 = (1 << 3)
    BLOCKS_8 = (1 << 2)
    BLOCKS_12 = (1 << 1)
    BLOCKS_16 = 1
    ALL = 0xF


class SBCChannelMode:
    """Indicate with which channel mode the frame has been
    encoded. The number of channels depends on this information."""
    CHANNEL_MODE_MONO = (1 << 3)
    CHANNEL_MODE_DUAL = (1 << 2)
    CHANNEL_MODE_STEREO = (1 << 1)
    CHANNEL_MODE_JOINT_STEREO = 1
    ALL = 0xF


class SBCAllocationMethod:
    """Indicates how the bit pool is allocated to different
    subbands. Either it is based on the loudness of the sub
    band signal or on the signal to noise ratio."""
    SNR = (1 << 1)
    LOUDNESS = 1
    ALL = 0x3


class SBCSubbands:
    """indicates the number of subbands with which the frame
    has been encoded"""
    SUBBANDS_4 = (1 << 1)
    SUBBANDS_8 = 1
    ALL = 0x3


class SBCCodec:
    """
    Python cass wrapper around CFFI calls into the SBC codec
    implemented in C.  The main API is defined by sbc.h with
    additional functions added to encapsulate an SBC payload
    as part of an RTP packet.  The API extensions for RTP
    are designed to work directly with bluetooth media
    transport file descriptors for the A2DP profile.  So,
    the basic idea is that this class is instantiated as
    part of a media endpoint implementation in order to
    encode or decode data carried on the media transport.

    .. note:: You need two separate instantiations of this
        class if you wish to encode and decode at the same
        time.  Although the class implementation is the same,
        the underlying C implementation requires separate
        `sbc_t` instances.

    :param namedtuple config: Media endpoint negotiated
        configuration parameters.  These are not used
        directly by the codec here but translated to
        parameters usable by the codec. See
        :py:class:`.SBCCodecConfig`
    """

    def __init__(self, config):

        import sys

        try:
            self.codec = ffi.verify(b'#include "rtpsbc.h"',
                                    libraries=[b'rtpsbc'],
                                    ext_package=b'rtpsbc')
        except:
            print 'Exception:', sys.exc_info()[0]

        self.config = ffi.new('sbc_t *')
        self.ts = ffi.new('unsigned int *', 0)
        self.seq_num = ffi.new('unsigned int *', 0)
        self._init_sbc_config(config)
        self.codec.sbc_init(self.config, 0)

    def _init_sbc_config(self, config):
        """
        Translator from namedtuple config representation to
        the sbc_t type.

        :param namedtuple config: See :py:class:`.SBCCodecConfig`
        :returns:
        """
        if (config.channel_mode == SBCChannelMode.CHANNEL_MODE_MONO):
            self.config.mode = self.codec.SBC_MODE_MONO
        elif (config.channel_mode == SBCChannelMode.CHANNEL_MODE_STEREO):
            self.config.mode = self.codec.SBC_MODE_STEREO
        elif (config.channel_mode == SBCChannelMode.CHANNEL_MODE_DUAL):
            self.config.mode = self.codec.SBC_MODE_DUAL_CHANNEL
        elif (config.channel_mode == SBCChannelMode.CHANNEL_MODE_JOINT_STEREO):
            self.config.mode = self.codec.SBC_MODE_JOINT_STEREO

        if (config.frequency == SBCSamplingFrequency.FREQ_16KHZ):
            self.config.frequency = self.codec.SBC_FREQ_16000
        elif (config.frequency == SBCSamplingFrequency.FREQ_32KHZ):
            self.config.frequency = self.codec.SBC_FREQ_32000
        elif (config.frequency == SBCSamplingFrequency.FREQ_44_1KHZ):
            self.config.frequency = self.codec.SBC_FREQ_44100
        elif (config.frequency == SBCSamplingFrequency.FREQ_48KHZ):
            self.config.frequency = self.codec.SBC_FREQ_48000

        if (config.allocation_method == SBCAllocationMethod.LOUDNESS):
            self.config.allocation = self.codec.SBC_AM_LOUDNESS
        elif (config.allocation_method == SBCAllocationMethod.SNR):
            self.config.allocation = self.codec.SBC_AM_SNR

        if (config.subbands == SBCSubbands.SUBBANDS_4):
            self.config.subbands = self.codec.SBC_SB_4
        elif (config.subbands == SBCSubbands.SUBBANDS_8):
            self.config.subbands = self.codec.SBC_SB_8

        if (config.block_length == SBCBlocks.BLOCKS_4):
            self.config.blocks = self.codec.SBC_BLK_4
        elif (config.block_length == SBCBlocks.BLOCKS_8):
            self.config.blocks = self.codec.SBC_BLK_8
        elif (config.block_length == SBCBlocks.BLOCKS_12):
            self.config.blocks = self.codec.SBC_BLK_12
        elif (config.block_length == SBCBlocks.BLOCKS_16):
            self.config.blocks = self.codec.SBC_BLK_16

        self.config.bitpool = config.max_bitpool
        self.config.endian = self.codec.SBC_LE

    def encode(self, fd, mtu, data):
        """
        Encode the supplied data (byte array) and write to
        the media transport file descriptor encapsulated
        as RTP packets.  The encoder will calculate the
        required number of SBC frames and encapsulate as
        RTP to fit the MTU size.

        :param int fd: Media transport file descriptor
        :param int mtu: Media transport MTU size as returned
            when the media transport was acquired.
        :param array{byte} data: Data to encode and send
            over the media transport.
        :return:
        """
        self.codec.rtp_sbc_encode_to_fd(self.config,
                                        ffi.new('char[]',
                                                data),
                                        len(data),
                                        mtu,
                                        self.ts,
                                        self.seq_num,
                                        fd)

    def decode(self, fd, mtu, max_len=2560):
        """
        Read the media transport descriptor, depay
        the RTP payload and decode the SBC frames into
        a byte array.  The maximum number of bytes to
        be returned may be passed as an argument and all
        available bytes are returned to the caller.

        :param int fd: Media transport file descriptor
        :param int mtu: Media transport MTU size as returned
            when the media transport was acquired.
        :param int max_len: Optional.  Set maximum number of
            bytes to read.
        :return data: Decoded data bytes as an array.
        :rtype: array{byte}
        """
        output_buffer = ffi.new('char[]', max_len)
        sz = self.codec.rtp_sbc_decode_from_fd(self.config,
                                               output_buffer,
                                               max_len,
                                               mtu,
                                               fd)
        return ffi.buffer(output_buffer[0:sz])



from __future__ import unicode_literals

VENDORS = {
    0x0000: 'Ericsson Technology Licensing',
    0x0001: 'Nokia Mobile Phones',
    0x0002: 'Intel Corp.',
    0x0003: 'IBM Corp.',
    0x0004: 'Toshiba Corp.',
    0x0005: '3Com',
    0x0006: 'Microsoft',
    0x0007: 'Lucent',
    0x0008: 'Motorola',
    0x0009: 'Infineon Technologies AG',
    0x000A: 'Cambridge Silicon Radio',
    0x000B: 'Silicon Wave',
    0x000C: 'Digianswer A/S',
    0x000D: 'Texas Instruments Inc.',
    0x000E: 'Ceva, Inc. (formerly Parthus Technologies, Inc.)',
    0x000F: 'Broadcom Corporation',
    0x0010: 'Mitel Semiconductor',
    0x0011: 'Widcomm, Inc',
    0x0012: 'Zeevo, Inc.',
    0x0013: 'Atmel Corporation',
    0x0014: 'Mitsubishi Electric Corporation',
    0x0015: 'RTX Telecom A/S',
    0x0016: 'KC Technology Inc.',
    0x0017: 'NewLogic',
    0x0018: 'Transilica, Inc.',
    0x0019: 'Rohde & Schwarz GmbH & Co. KG',
    0x001A: 'TTPCom Limited',
    0x001B: 'Signia Technologies, Inc.',
    0x001C: 'Conexant Systems Inc.',
    0x001D: 'Qualcomm',
    0x001E: 'Inventel',
    0x001F: 'AVM Berlin',
    0x0020: 'BandSpeed, Inc.',
    0x0021: 'Mansella Ltd',
    0x0022: 'NEC Corporation',
    0x0023: 'WavePlus Technology Co., Ltd.',
    0x0024: 'Alcatel',
    0x0025: 'Philips Semiconductors',
    0x0026: 'C Technologies',
    0x0027: 'Open Interface',
    0x0028: 'R F Micro Devices',
    0x0029: 'Hitachi Ltd',
    0x002A: 'Symbol Technologies, Inc.',
    0x002B: 'Tenovis',
    0x002C: 'Macronix International Co. Ltd.',
    0x002D: 'GCT Semiconductor',
    0x002E: 'Norwood Systems',
    0x002F: 'MewTel Technology Inc.',
    0x0030: 'ST Microelectronics',
    0x0031: 'Synopsis',
    0x0032: 'Red-M (Communications) Ltd',
    0x0033: 'Commil Ltd',
    0x0034: 'Computer Access Technology Corporation (CATC)',
    0x0035: 'Eclipse (HQ Espana) S.L.',
    0x0036: 'Renesas Technology Corp.',
    0x0037: 'Mobilian Corporation',
    0x0038: 'Terax',
    0x0039: 'Integrated System Solution Corp.',
    0x003A: 'Matsushita Electric Industrial Co., Ltd.',
    0x003B: 'Gennum Corporation',
    0x003C: 'Research In Motion',
    0x003D: 'IPextreme, Inc.',
    0x003E: 'Systems and Chips, Inc.',
    0x003F: 'Bluetooth SIG, Inc.',
    0x0040: 'Seiko Epson Corporation',
    0x0041: 'Integrated Silicon Solution Taiwan, Inc.',
    0x0042: 'CONWISE Technology Corporation Ltd',
    0x0043: 'PARROT SA',
    0x0044: 'Socket Mobile',
    0x0045: 'Atheros Communications, Inc.',
    0x0046: 'MediaTek, Inc.',
    0x0047: 'Bluegiga',
    0x0048: 'Marvell Technology Group Ltd.',
    0x0049: '3DSP Corporation',
    0x004A: 'Accel Semiconductor Ltd.',
    0x004B: 'Continental Automotive Systems',
    0x004C: 'Apple, Inc.',
    0x004D: 'Staccato Communications, Inc.',
    0x004E: 'Avago Technologies',
    0x004F: 'APT Licensing Ltd.',
    0x0050: 'SiRF Technology',
    0x0051: 'Tzero Technologies, Inc.',
    0x0052: 'J&M Corporation',
    0x0053: 'Free2move AB',
    0x0054: '3DiJoy Corporation',
    0x0055: 'Plantronics, Inc.',
    0x0056: 'Sony Ericsson Mobile Communications',
    0x0057: 'Harman International Industries, Inc.',
    0x0058: 'Vizio, Inc.',
    0x0059: 'Nordic Semiconductor ASA',
    0x005A: 'EM Microelectronic-Marin SA',
    0x005B: 'Ralink Technology Corporation',
    0x005C: 'Belkin International, Inc.',
    0x005D: 'Realtek Semiconductor Corporation',
    0x005E: 'Stonestreet One, LLC',
    0x005F: 'Wicentric, Inc.',
    0x0060: 'RivieraWaves S.A.S',
    0x0061: 'RDA Microelectronics',
    0x0062: 'Gibson Guitars',
    0x0063: 'MiCommand Inc.',
    0x0064: 'Band XI International, LLC',
    0x0065: 'Hewlett-Packard Company',
    0x0066: '9Solutions Oy',
    0x0067: 'GN Netcom A/S',
    0x0068: 'General Motors',
    0x0069: 'A&D Engineering, Inc.',
    0x006A: 'MindTree Ltd.',
    0x006B: 'Polar Electro OY',
    0x006C: 'Beautiful Enterprise Co., Ltd.',
    0x006D: 'BriarTek, Inc.',
    0x006E: 'Summit Data Communications, Inc.',
    0x006F: 'Sound ID',
    0x0070: 'Monster, LLC',
    0x0071: 'connectBlue AB',
    0x0072: 'ShangHai Super Smart Electronics Co. Ltd.',
    0x0073: 'Group Sense Ltd.',
    0x0074: 'Zomm, LLC',
    0x0075: 'Samsung Electronics Co. Ltd.',
    0x0076: 'Creative Technology Ltd.',
    0x0077: 'Laird Technologies',
    0x0078: 'Nike, Inc.',
    0x0079: 'lesswire AG',
    0x007A: 'MStar Semiconductor, Inc.',
    0x007B: 'Hanlynn Technologies',
    0x007C: 'A & R Cambridge',
    0x007D: 'Seers Technology Co. Ltd',
    0x007E: 'Sports Tracking Technologies Ltd.',
    0x007F: 'Autonet Mobile',
    0x0080: 'DeLorme Publishing Company, Inc.',
    0x0081: 'WuXi Vimicro',
    0x0082: 'Sennheiser Communications A/S',
    0x0083: 'TimeKeeping Systems, Inc.',
    0x0084: 'Ludus Helsinki Ltd.',
    0x0085: 'BlueRadios, Inc.',
    0x0086: 'equinox AG',
    0x0087: 'Garmin International, Inc.',
    0x0088: 'Ecotest',
    0x0089: 'GN ReSound A/S',
    0x008A: 'Jawbone',
    0x008B: 'Topcorn Positioning Systems, LLC',
    0x008C: 'Qualcomm Retail Solutions, Inc. (formerly Qualcomm Labs, Inc.)',
    0x008D: 'Zscan Software',
    0x008E: 'Quintic Corp.',
    0x008F: 'Stollman E+V GmbH',
    0x0090: 'Funai Electric Co., Ltd.',
    0x0091: 'Advanced PANMOBIL Systems GmbH & Co. KG',
    0x0092: 'ThinkOptics, Inc.',
    0x0093: 'Universal Electronics, Inc.',
    0x0094: 'Airoha Technology Corp.',
    0x0095: 'NEC Lighting, Ltd.',
    0x0096: 'ODM Technology, Inc.',
    0x0097: 'ConnecteDevice Ltd.',
    0x0098: 'zer01.tv GmbH',
    0x0099: 'i.Tech Dynamic Global Distribution Ltd.',
    0x009A: 'Alpwise',
    0x009B: 'Jiangsu Toppower Automotive Electronics Co., Ltd.',
    0x009C: 'Colorfy, Inc.',
    0x009D: 'Geoforce Inc.',
    0x009E: 'Bose Corporation',
    0x009F: 'Suunto Oy',
    0x00A0: 'Kensington Computer Products Group',
    0x00A1: 'SR-Medizinelektronik',
    0x00A2: 'Vertu Corporation Limited',
    0x00A3: 'Meta Watch Ltd.',
    0x00A4: 'LINAK A/S',
    0x00A5: 'OTL Dynamics LLC',
    0x00A6: 'Panda Ocean Inc.',
    0x00A7: 'Visteon Corporation',
    0x00A8: 'ARP Devices Limited',
    0x00A9: 'Magneti Marelli S.p.A',
    0x00AA: 'CAEN RFID srl',
    0x00AB: 'Ingenieur-Systemgruppe Zahn GmbH',
    0x00AC: 'Green Throttle Games',
    0x00AD: 'Peter Systemtechnik GmbH',
    0x00AE: 'Omegawave Oy',
    0x00AF: 'Cinetix',
    0x00B0: 'Passif Semiconductor Corp',
    0x00B1: 'Saris Cycling Group, Inc',
    0x00B2: 'Bekey A/S',
    0x00B3: 'Clarinox Technologies Pty. Ltd.',
    0x00B4: 'BDE Technology Co., Ltd.',
    0x00B5: 'Swirl Networks',
    0x00B6: 'Meso international',
    0x00B7: 'TreLab Ltd',
    0x00B8: 'Qualcomm Innovation Center, Inc. (QuIC)',
    0x00B9: 'Johnson Controls, Inc.',
    0x00BA: 'Starkey Laboratories Inc.',
    0x00BB: 'S-Power Electronics Limited',
    0x00BC: 'Ace Sensor Inc',
    0x00BD: 'Aplix Corporation',
    0x00BE: 'AAMP of America',
    0x00BF: 'Stalmart Technology Limited',
    0x00C0: 'AMICCOM Electronics Corporation',
    0x00C1: 'Shenzhen Excelsecu Data Technology Co.,Ltd',
    0x00C2: 'Geneq Inc.',
    0x00C3: 'adidas AG',
    0x00C4: 'LG Electronics',
    0x00C5: 'Onset Computer Corporation',
    0x00C6: 'Selfly BV',
    0x00C7: 'Quuppa Oy.',
    0x00C8: 'GeLo Inc',
    0x00C9: 'Evluma',
    0x00CA: 'MC10',
    0x00CB: 'Binauric SE',
    0x00CC: 'Beats Electronics',
    0x00CD: 'Microchip Technology Inc.',
    0x00CE: 'Elgato Systems GmbH',
    0x00CF: 'ARCHOS SA',
    0x00D0: 'Dexcom, Inc.',
    0x00D1: 'Polar Electro Europe B.V.',
    0x00D2: 'Dialog Semiconductor B.V.',
    0x00D3: 'Taixingbang Technology (HK) Co,. LTD.',
    0x00D4: 'Kawantech',
    0x00D5: 'Austco Communication Systems',
    0x00D6: 'Timex Group USA, Inc.',
    0x00D7: 'Qualcomm Technologies, Inc.',
    0x00D8: 'Qualcomm Connected Experiences, Inc.',
    0x00D9: 'Voyetra Turtle Beach',
    0x00DA: 'txtr GmbH',
    0x00DB: 'Biosentronics',
    0x00DC: 'Procter & Gamble',
    0x00DD: 'Hosiden Corporation',
    0x00DE: 'Muzik LLC',
    0x00DF: 'Misfit Wearables Corp',
    0x00E0: 'Google',
    0x00E1: 'Danlers Ltd',
    0x00E2: 'Semilink Inc',
    0x00E3: 'inMusic Brands, Inc',
    0x00E4: 'L.S. Research Inc.',
    0x00E5: 'Eden Software Consultants Ltd.',
    0x00E6: 'Freshtemp',
    0x00E7: 'KS Technologies',
    0x00E8: 'ACTS Technologies',
    0x00E9: 'Vtrack Systems',
    0x00EA: 'Nielsen-Kellerman Company',
    0x00EB: 'Server Technology, Inc.',
    0x00EC: 'BioResearch Associates',
    0x00ED: 'Jolly Logic, LLC',
    0x00EE: 'Above Average Outcomes, Inc.',
    0x00EF: 'Bitsplitters GmbH',
    0x00F0: 'PayPal, Inc.',
    0x00F1: 'Witron Technology Limited',
    0x00F2: 'Aether Things Inc. (formerly Morse Project Inc.)',
    0x00F3: 'Kent Displays Inc.',
    0x00F4: 'Nautilus Inc.',
    0x00F5: 'Smartifier Oy',
    0x00F6: 'Elcometer Limited',
    0x00F7: 'VSN Technologies Inc.',
    0x00F8: 'AceUni Corp., Ltd.',
    0x00F9: 'StickNFind',
    0x00FA: 'Crystal Code AB',
    0x00FB: 'KOUKAAM a.s.',
    0x00FC: 'Delphi Corporation',
    0x00FD: 'ValenceTech Limited',
    0x00FE: 'Reserved',
    0x00FF: 'Typo Products, LLC',
    0x0100: 'TomTom International BV',
    0x0101: 'Fugoo, Inc',
    0x0102: 'Keiser Corporation',
    0x0103: 'Bang & Olufsen A/S',
    0x0104: 'PLUS Locations Systems Pty Ltd',
    0x0105: 'Ubiquitous Computing Technology Corporation',
    0x0106: 'Innovative Yachtter Solutions',
    0x0107: 'William Demant Holding A/S',
    0x0108: 'Chicony Electronics Co., Ltd.',
    0x0109: 'Atus BV',
    0x010A: 'Codegate Ltd.',
    0x010B: 'ERi, Inc.',
    0x010C: 'Transducers Direct, LLC',
    0x010D: 'Fujitsu Ten Limited',
    0x010E: 'Audi AG',
    0x010F: 'HiSilicon Technologies Co., Ltd.',
    0x0110: 'Nippon Seiki Co., Ltd.',
    0x0111: 'Steelseries ApS',
    0x0112: 'vyzybl Inc.',
    0x0113: 'Openbrain Technologies, Co., Ltd.',
    0x0114: 'Xensr',
    0x0115: 'e.solutions',
    0x0116: '1OAK Technologies',
    0x0117: 'Wimoto Technologies Inc',
    0x0118: 'Radius Networks, Inc.',
    0x0119: 'Wize Technology Co., Ltd.',
    0x011A: 'Qualcomm Labs, Inc.',
    0x011B: 'Aruba Networks',
    0x011C: 'Baidu',
    0x011D: 'Arendi AG',
    0x011E: 'Skoda Auto a.s.',
    0x011F: 'Volkswagon AG',
    0x0120: 'Porsche AG',
    0x0121: 'Sino Wealth Electronic Ltd.',
    0x0122: 'AirTurn, Inc.',
    0x0123: 'Kinsa, Inc.',
    0x0124: 'HID Global',
    0x0125: 'SEAT es',
    0x0126: 'Promethean Ltd.',
    0x0127: 'Salutica Allied Solutions',
    0x0128: 'GPSI Group Pty Ltd',
    0x0129: 'Nimble Devices Oy',
    0x012A: 'Changzhou Yongse Infotech Co., Ltd',
    0x012B: 'SportIQ',
    0x012C: 'TEMEC Instruments B.V.',
    0x012D: 'Sony Corporation',
    0x012E: 'ASSA ABLOY',
    0x012F: 'Clarion Co., Ltd.',
    0x0130: 'Warehouse Innovations',
    0x0131: 'Cypress Semiconductor Corporation',
    0x0132: 'MADS Inc',
    0x0133: 'Blue Maestro Limited',
    0x0134: 'Resolution Products, Inc.',
    0x0135: 'Airewear LLC',
    0x0136: 'ETC sp. z.o.o.',
    0x0137: 'Prestigio Plaza Ltd.',
    0x0138: 'NTEO Inc.',
    0x0139: 'Focus Systems Corporation',
    0x013A: 'Tencent Holdings Limited',
    0x013B: 'Allegion',
    0x013C: 'Murata Manufacuring Co., Ltd.',
    0x013D: 'WirelessWERX',
    0x013E: 'Nod, Inc.',
    0x013F: 'B&B Manufacturing Company',
    0x0140: 'Alpine Electronics (China) Co., Ltd',
    0x0141: 'FedEx Services',
    0x0142: 'Grape Systems Inc.',
    0x0143: 'Bkon Connect',
    0x0144: 'Lintech GmbH',
    0x0145: 'Novatel Wireless',
    0x0146: 'Ciright',
    0x0147: 'Mighty Cast, Inc.',
    0x0148: 'Ambimat Electronics',
    0x0149: 'Perytons Ltd.',
    0x014A: 'Tivoli Audio, LLC',
}
